import pytest
from unittest.mock import Mock

from pynonymizer.strategy.database import DatabaseStrategy
from pynonymizer.strategy.table import TruncateTableStrategy, UpdateColumnsTableStrategy, DeleteTableStrategy
from pynonymizer.strategy.update_column import UniqueEmailUpdateColumnStrategy, UniqueLoginUpdateColumnStrategy, FakeUpdateColumnStrategy, EmptyUpdateColumnStrategy
from pynonymizer.fake import FakeDataType


@pytest.fixture
def simple_strategy_fake_generator():
    return Mock(get_data_type=Mock(return_value=FakeDataType.STRING), get_value=Mock(return_value="TEST_VALUE"))


@pytest.fixture
def simple_strategy_update_fake_column(simple_strategy_fake_generator):
    return FakeUpdateColumnStrategy("column3", simple_strategy_fake_generator, "user_name", where="BANANAS < 3")


@pytest.fixture
def simple_strategy_update(simple_strategy_update_fake_column):
    return UpdateColumnsTableStrategy("update_table_where_3", [
            UniqueEmailUpdateColumnStrategy("column1", where="BANANAS < 5"),
            UniqueLoginUpdateColumnStrategy("column2", where="BANANAS < 5"),
            simple_strategy_update_fake_column,
            EmptyUpdateColumnStrategy("column4")
        ])

@pytest.fixture
def simple_strategy_delete():
    return DeleteTableStrategy("delete_table")

@pytest.fixture
def simple_strategy_trunc():
    return TruncateTableStrategy("truncate_table")

@pytest.fixture
def simple_strategy_schema_trunc():
    return TruncateTableStrategy("truncate_schema_table", schema="schema")


@pytest.fixture
def simple_strategy(simple_strategy_trunc, simple_strategy_update, simple_strategy_delete):
    return DatabaseStrategy([
        simple_strategy_trunc,
        simple_strategy_update,
        simple_strategy_delete
    ])




from abc import ABC, abstractmethod

SEED_TABLE_NAME = "_pynonymizer_seed_fake_data"

class DatabaseProvider(ABC):
    @abstractmethod
    def __init__(self, db_host, db_user, db_pass, db_name, db_port=None, seed_rows=None):
        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.db_port = db_port

        if seed_rows is None:
            seed_rows = 150

        self.seed_rows = int(seed_rows)

    @abstractmethod
    def create_database(self):
        pass

    @abstractmethod
    def drop_database(self):
        pass

    @abstractmethod
    def anonymize_database(self, database_strategy):
        pass

    @abstractmethod
    def restore_database(self, input_obj):
        pass

    @abstractmethod
    def dump_database(self, output_obj):
        pass



import unittest
from unittest.mock import Mock

from pynonymizer.fake import UnsupportedFakeTypeError
from pynonymizer.strategy import database
from pynonymizer.strategy.exceptions import UnknownTableStrategyError, UnknownColumnStrategyError, ConfigSyntaxError
from pynonymizer.strategy.table import TableStrategyTypes
from pynonymizer.strategy.update_column import UpdateColumnStrategyTypes
import copy
import pytest


@pytest.fixture
def simple_config():
    return {
        "tables": {
            "accounts": {
                "columns": {
                    "current_sign_in_ip": "ipv4_public",
                    "username": "unique_login",
                    "email": "unique_email",
                    "name": "empty",
                }
            },
            "secrets" : "delete",

            "transactions": "truncate"
        }
    }

@pytest.fixture
def config_unsupported_fake_type():
    return {
            "tables": {
                "accounts": {
                    "columns": {
                        "current_sign_in_ip": "ipv4_public",
                        "username": "unique_login",
                        "email": "NOT A VALID FAKE TYPE",
                        "name": "empty",
                    }
                },

                "transactions": "truncate"
            }
        }

@pytest.fixture
def fake_column_generator():
    from pynonymizer.fake import FakeColumnGenerator
    return FakeColumnGenerator()


@pytest.fixture
def strategy_parser(fake_column_generator):
    from pynonymizer.strategy import parser
    return parser.StrategyParser(fake_column_generator)


def test_valid_parse_no_mutate(simple_config, strategy_parser):
    old_valid_config = copy.deepcopy(simple_config)
    strategy_parser.parse_config(simple_config)

    assert simple_config == old_valid_config

def test_simple_parse_creates_databasestrategy(simple_config, strategy_parser):
    strategy = strategy_parser.parse_config(simple_config)
    
    assert isinstance(strategy, database.DatabaseStrategy)

def test_simple_parse_update_columns(simple_config, strategy_parser):
    strategy = strategy_parser.parse_config(simple_config)

    table = strategy.table_strategies[0]
    assert table.table_name == "accounts"
    assert table.strategy_type == TableStrategyTypes.UPDATE_COLUMNS


def test_simple_parse_truncate(simple_config, strategy_parser):
    strategy = strategy_parser.parse_config(simple_config)

    table = strategy.table_strategies[2]
    assert table.table_name == "transactions"
    assert table.strategy_type == TableStrategyTypes.TRUNCATE

def test_simple_parse_delete(simple_config, strategy_parser):
    strategy = strategy_parser.parse_config(simple_config)

    table = strategy.table_strategies[1]
    assert table.table_name == "secrets"
    assert table.strategy_type == TableStrategyTypes.DELETE


    # need a better matching technique than checking list indicies.
    """
    assert accounts_strategy.column_strategies[0].strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE

    assert accounts_strategy.column_strategies[1].strategy_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN
    assert accounts_strategy.column_strategies[2].strategy_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL
    assert accounts_strategy.column_strategies[3].strategy_type == UpdateColumnStrategyTypes.EMPTY
    """



def test_unsupported_fake_column_type(strategy_parser, config_unsupported_fake_type):
    """
    get_fake_column's UnsupportedFakeType should kill a parse attempt
    """

    with pytest.raises(UnsupportedFakeTypeError):
        strategy_parser.parse_config(config_unsupported_fake_type)


def test_invalid_table_strategy_parse(strategy_parser):
    with pytest.raises(UnknownTableStrategyError):
        strategy_parser.parse_config({
            "tables": {
                "accounts": "cheesecake"
            }
        })


def test_unknown_column_strategy(strategy_parser):
    with pytest.raises(UnknownColumnStrategyError):
        strategy_parser.parse_config({
            "tables": {
                "accounts": {
                    "columns": {
                        "current_sign_in_ip": {
                            "type": "cheese"  # Not a valid strategy
                        }
                    }
                },

                "transactions": "truncate"
            }
        })


def test_unknown_table_strategy_bad_dict(strategy_parser):
    with pytest.raises(UnknownTableStrategyError):
        strategy_parser.parse_config({
            "tables": {
                "accounts": {
                    "not_columns": {
                        "current_sign_in_ip": "ipv4_public",
                        "username": "unique_login",
                        "email": "unique_email",
                        "name": "empty",
                    }
                },
            }
        })


def test_valid_parse_before_after_script(strategy_parser):
    parse_result = strategy_parser.parse_config({
        "scripts": {
            "before": [
                "SELECT `before` from `students`;"
            ],
            "after": [
                "SELECT `after` from `students`;",
                "SELECT `after_2` from `example`;"
            ]
        },
        "tables": {
            "accounts": "truncate"
        },
    })

    assert isinstance(parse_result, database.DatabaseStrategy)

    assert len(parse_result.table_strategies) == 1
    assert parse_result.table_strategies[0].strategy_type == TableStrategyTypes.TRUNCATE

    assert parse_result.scripts["before"] == [
                "SELECT `before` from `students`;"
            ]
    assert parse_result.scripts["after"] == [
                "SELECT `after` from `students`;",
                "SELECT `after_2` from `example`;"
            ]


def test_verbose_table_truncate(strategy_parser):
    with pytest.raises(ConfigSyntaxError):
        strategy = strategy_parser.parse_config({
            "tables": {
                "table1": {
                    "type": "truncate",
                    # parser should raise error when keys from other types when type is specified
                    "columns": {}
                }
            }
        })


def test_verbose_table_update_columns(strategy_parser):
    strategy = strategy_parser.parse_config({
        "tables": {
            "table1": {
                "type": "update_columns",
                "columns": {
                }
            }
        }
    })

    assert len(strategy.table_strategies) == 1
    assert strategy.table_strategies[0].table_name == "table1"
    assert strategy.table_strategies[0].strategy_type == TableStrategyTypes.UPDATE_COLUMNS


def test_verbose_table_list_duplicate(strategy_parser):
    """parser should allow multiple tables of the same name in list-parse-mode"""
    strategy = strategy_parser.parse_config({
        "tables": [
            {
                "table_name": "table1",
                "type": "truncate",
            },
            {
                "table_name": "table1",
                "type": "truncate",
            }
        ]
    })

    assert len(strategy.table_strategies) == 2
    assert strategy.table_strategies[0].table_name == "table1"
    assert strategy.table_strategies[0].strategy_type == TableStrategyTypes.TRUNCATE
    assert strategy.table_strategies[1].table_name == "table1"
    assert strategy.table_strategies[1].strategy_type == TableStrategyTypes.TRUNCATE


def test_table_raises_when_given_unrelated_key(strategy_parser):
    with pytest.raises(ConfigSyntaxError):
        strategy_parser.parse_config({
            "tables": {
                "table1": {
                    "type": "update_columns",
                    "columns": {
                        "column1": {
                            "type": "empty",
                            "where": "condition = 'value1'",
                            "fake_type": "email"
                        },
                    }
                }
            }
        })


import re
import pytest
from contextlib import contextmanager


class AnyObject:
    def __eq__(self, actual):
        return True

    def __ne__(self, other):
        return False


class SuperdictOf:
    def __init__(self, required_dict):
        self.required_dict = required_dict

    def __eq__(self, actual):
        return self.required_dict.items() <= actual.items()

    def __ne__(self, actual):
        return not(self.required_dict.items() <= actual.items())



class ComparableRegex:
    """Assert that a given string meets some expectations."""

    def __init__(self, pattern, flags=0):
        self._regex = re.compile(pattern, flags)

    def __eq__(self, actual):
        return bool(self._regex.match(actual))

    def __repr__(self):
        return self._regex.pattern


@contextmanager
def not_raises(exception):
    try:
        yield
    except exception:
        raise pytest.fail("DID RAISE {0}".format(exception))


def list_rindex(alist, value):
  return len(alist) - alist[-1::-1].index(value) - 1


import unittest
from unittest.mock import Mock
import pytest
import re
from pynonymizer.fake import FakeDataType
from pynonymizer.database.exceptions import UnsupportedColumnStrategyError
from pynonymizer.strategy.database import DatabaseStrategy
from pynonymizer.strategy.table import UpdateColumnsTableStrategy
from pynonymizer.strategy.update_column import (
    FakeUpdateColumnStrategy,
    EmptyUpdateColumnStrategy,
    UniqueLoginUpdateColumnStrategy,
    UniqueEmailUpdateColumnStrategy,
    LiteralUpdateColumnStrategy
)
import pynonymizer.database.mysql.query_factory as query_factory

"""
These tests are brittle and based on the actual SQL generatedColumnStrategyTypes. 
The sentiment is to test the 'meaning' of the SQL, rather than the actual formatting, so it may be prudent to replace
these tests with some form of parsing or pattern matching. 

The general idea, however, is that by keeping the queryfactory separate from the provider, it will not change often,
and the sql returned should be very stable.
"""


def test_get_truncate_table():
    assert query_factory.get_truncate_table("test") == "SET FOREIGN_KEY_CHECKS=0; " \
           "TRUNCATE TABLE `test`; SET FOREIGN_KEY_CHECKS=1;"

def test_get_delete_table():
    assert query_factory.get_delete_table("test") == "DELETE FROM `test`;"

def test_get_drop_seed_table():
    assert query_factory.get_drop_seed_table("seed_table") == "DROP TABLE IF EXISTS `seed_table`;"


def test_get_create_database():
    assert query_factory.get_create_database("test_database") == "CREATE DATABASE `test_database`;"


def test_get_drop_database():
    assert query_factory.get_drop_database("test_database") == "DROP DATABASE IF EXISTS `test_database`;"


def test_get_dumpsize_estimate():
    assert query_factory.get_dumpsize_estimate("test") == "SELECT data_bytes FROM " \
           "(SELECT SUM(data_length) AS data_bytes " \
           "FROM information_schema.tables WHERE table_schema = 'test') AS data;"

@pytest.fixture
def str_fake_column_generator():
    return Mock(
            get_data_type=Mock(return_value=FakeDataType.STRING),
            get_value=Mock(return_value="test_value")
        )
@pytest.fixture
def int_fake_column_generator():
    return Mock(
        get_data_type=Mock(return_value=FakeDataType.INT),
        get_value=Mock(return_value=645)
    )


@pytest.fixture
def fake_update_column_str_first_name(str_fake_column_generator):
    return FakeUpdateColumnStrategy("test_column1",str_fake_column_generator, "first_name")


@pytest.fixture
def fake_update_column_int_last_name(int_fake_column_generator):
    return FakeUpdateColumnStrategy("test_column2",int_fake_column_generator, "last_name")


@pytest.fixture
def empty_strategy():
    return EmptyUpdateColumnStrategy("test_column3")


@pytest.fixture
def ulogin_strategy():
    return UniqueLoginUpdateColumnStrategy("test_column4")


@pytest.fixture
def uemail_strategy():
    return UniqueEmailUpdateColumnStrategy("test_column5")


@pytest.fixture
def literal_strategy():
    return LiteralUpdateColumnStrategy("test_column6", value="RAND()")


@pytest.fixture
def database_strategy():
    pass


@pytest.fixture
def qualifier_column_map(fake_update_column_str_first_name,fake_update_column_int_last_name,empty_strategy,ulogin_strategy, uemail_strategy,literal_strategy):
    return {
        "first_name": fake_update_column_str_first_name,
        "last_name": fake_update_column_int_last_name,
        "first_name_test_arg_5": fake_update_column_str_first_name
    }


@pytest.fixture
def column_strategy_list(fake_update_column_str_first_name,fake_update_column_int_last_name,empty_strategy,ulogin_strategy, uemail_strategy,literal_strategy):
    return [
        fake_update_column_str_first_name,
        fake_update_column_int_last_name,
        empty_strategy,
        ulogin_strategy,
        uemail_strategy,
        literal_strategy
    ]


@pytest.fixture
def update_table_strategy(column_strategy_list):
    return UpdateColumnsTableStrategy("table_name", column_strategy_list)


@pytest.fixture
def update_table_strategy_unknown(unsupported_column_strategy):
    return UpdateColumnsTableStrategy("invalid_table", [unsupported_column_strategy])


@pytest.fixture
def unsupported_column_strategy():
    column = UniqueLoginUpdateColumnStrategy("column_name")
    column.strategy_type = "NOT_SUPPORTED"

    return column


def test_get_insert_seed_row(qualifier_column_map):
    insert_seed_row = query_factory.get_insert_seed_row("seed_table", qualifier_column_map)

    assert insert_seed_row == "INSERT INTO `seed_table`(`first_name`,`last_name`,`first_name_test_arg_5`) " \
                              "VALUES ('test_value',645,'test_value');"


def test_get_create_seed_table(qualifier_column_map):
    assert query_factory.get_create_seed_table("seed_table", qualifier_column_map) == "CREATE TABLE `seed_table` (`first_name` TEXT,`last_name` INT,`first_name_test_arg_5` TEXT);"


def test_get_create_seed_table_no_columns():
    """
    get_create_seed_table should error when presented with no columns
    """
    with pytest.raises(ValueError) as e_info:
        query_factory.get_create_seed_table("seed_table", {})


def test_get_update_table_unsupported_column_type(update_table_strategy_unknown):
    """
    get_update_table should raise UnsupportedColumnStrategyError if presented with an unsupported column type
    """
    with pytest.raises(UnsupportedColumnStrategyError):
        query_factory.get_update_table("seed_table", update_table_strategy_unknown)


def test_get_update_table_fake_column(column_strategy_list):
    update_table_all = query_factory.get_update_table("seed_table", UpdateColumnsTableStrategy("anon_table", column_strategy_list))

    assert update_table_all == [
            "UPDATE `anon_table` SET "
            "`test_column1` = ( SELECT `first_name` FROM `seed_table` ORDER BY RAND() LIMIT 1),"
            "`test_column2` = ( SELECT `last_name` FROM `seed_table` ORDER BY RAND() LIMIT 1),"
            "`test_column3` = (''),"
            "`test_column4` = ( SELECT MD5(FLOOR((NOW() + RAND()) * (RAND() * RAND() / RAND()) + RAND())) ),"
            "`test_column5` = ( SELECT CONCAT(MD5(FLOOR((NOW() + RAND()) * (RAND() * RAND() / RAND()) + RAND())), '@', MD5(FLOOR((NOW() + RAND()) * (RAND() * RAND() / RAND()) + RAND())), '.com') ),"
            "`test_column6` = RAND();"
            ]


def test_get_update_table_literal(literal_strategy):

    result_queries = query_factory.get_update_table("seed_table", UpdateColumnsTableStrategy("anon_table", [
        LiteralUpdateColumnStrategy("literal_column", "RAND()")
    ]))

    assert result_queries == [
        "UPDATE `anon_table` SET `literal_column` = RAND();"
    ]